@@ -19,12 +19,14 @@
 
 #include <config.h>
 #include <glib.h>
+#include <glib/gstdio.h>
 #include <string.h>
 #include <gtk/gtk.h>
 
 #include <gconf/gconf-client.h>
 
 #include <libgnome/libgnome.h>
+#include <libgnome/gnome-desktop-item.h>
 #include <libgnomeui/gnome-client.h>
 
 #include "save.h"
@@ -182,7 +184,7 @@ write_session (void)
   for (list = pending_list; list; list = list->next)
     {
       Client *client = (Client *) list->data;
-      
+
       g_snprintf (prefix, sizeof(prefix),
 		  CONFIG_PREFIX "%s/%d,", session_name, i);
 
@@ -194,7 +196,7 @@ write_session (void)
   for (list = purge_drop_list; list; list = list->next)
     {
       Client *client = (Client *) list->data;
-      
+
       g_snprintf (prefix, sizeof(prefix),
 		  CONFIG_PREFIX "%s/%d,", session_name, i);
       gnome_config_push_prefix (prefix);
@@ -205,7 +207,7 @@ write_session (void)
   for (list = purge_retain_list; list; list = list->next)
     {
       Client *client = (Client *) list->data;
-      
+
       g_snprintf (prefix, sizeof(prefix),
 		  CONFIG_PREFIX "%s/%d,", session_name, i);
       gnome_config_push_prefix (prefix);
@@ -355,6 +357,216 @@ read_clients (const char* file, const char *session, MatchRule match_rule)
   return list;
 }
 
+/* read .desktop files in a given directory */
+static void
+read_desktop_entries_in_dir (GHashTable *clients, const gchar *path)
+{
+  GDir *dir;
+  const gchar *name;
+
+  dir = g_dir_open (path, 0, NULL);
+  if (!dir)
+    return;
+
+  while ((name = g_dir_read_name (dir)))
+    {
+      Client *client = NULL;
+      Client *hash_client;
+      SmProp *prop;
+      gchar *desktop_file;
+      GnomeDesktopItem *ditem;
+
+      if (!g_str_has_suffix (name, ".desktop"))
+        continue;
+
+      desktop_file = g_build_filename (path, name, NULL);
+      ditem = gnome_desktop_item_new_from_file (desktop_file, GNOME_DESKTOP_ITEM_LOAD_NO_TRANSLATIONS, NULL);
+      if (ditem != NULL)
+        {
+	  const gchar *exec_string;
+	  gchar **argv, *hash_key;
+	  gint argc, i;
+
+	  exec_string = gnome_desktop_item_get_string (ditem, GNOME_DESKTOP_ITEM_EXEC);
+          /* First filter out entries without Exec keys and hidden entries */
+	  if ((exec_string == NULL) || 
+              !g_shell_parse_argv (exec_string, &argc, &argv, NULL) ||
+              (gnome_desktop_item_attr_exists (ditem, GNOME_DESKTOP_ITEM_HIDDEN) &&
+               gnome_desktop_item_get_boolean (ditem, GNOME_DESKTOP_ITEM_HIDDEN)))
+            {
+              gnome_desktop_item_unref (ditem);
+              g_free (desktop_file);
+              continue;
+            }
+
+          /* The filter out entries that are not for GNOME */
+          if (gnome_desktop_item_attr_exists (ditem, GNOME_DESKTOP_ITEM_ONLY_SHOW_IN))
+              {
+                int i;
+                char **only_show_in_list;
+
+                only_show_in_list = 
+                    gnome_desktop_item_get_strings (ditem,
+                                                    GNOME_DESKTOP_ITEM_ONLY_SHOW_IN);
+
+                for (i = 0; only_show_in_list[i] != NULL; i++)
+                  {
+                    if (g_strcasecmp (only_show_in_list[i], "GNOME") == 0)
+                      break;
+                  }
+
+                if (only_show_in_list[i] == NULL)
+                  {
+                    gnome_desktop_item_unref (ditem);
+                    g_free (desktop_file);
+                    g_strfreev (argv);
+                    g_strfreev (only_show_in_list);
+                    continue;
+                  }
+                g_strfreev (only_show_in_list);
+              }
+          if (gnome_desktop_item_attr_exists (ditem, "NotShowIn"))
+              {
+                int i;
+                char **not_show_in_list;
+
+                not_show_in_list = 
+                    gnome_desktop_item_get_strings (ditem, "NotShowIn");
+
+                for (i = 0; not_show_in_list[i] != NULL; i++)
+                  {
+                    if (g_strcasecmp (not_show_in_list[i], "GNOME") == 0)
+                      break;
+                  }
+
+                if (not_show_in_list[i] != NULL)
+                  {
+                    gnome_desktop_item_unref (ditem);
+                    g_free (desktop_file);
+                    g_strfreev (argv);
+                    g_strfreev (not_show_in_list);
+                    continue;
+                  }
+                g_strfreev (not_show_in_list);
+              }
+
+          /* finally filter out entries that require a program that's not
+           * installed 
+           */
+          if (gnome_desktop_item_attr_exists (ditem,
+                                              GNOME_DESKTOP_ITEM_TRY_EXEC))
+            {
+              const char *program;
+              char *program_path;
+              program = gnome_desktop_item_get_string (ditem,
+                                                       GNOME_DESKTOP_ITEM_TRY_EXEC);
+              program_path = g_find_program_in_path (program);
+              if (!program_path)
+                {
+                  gnome_desktop_item_unref (ditem);
+                  g_free (desktop_file);
+                  g_strfreev (argv);
+                  continue;
+                }
+              g_free (program_path);
+            }
+
+	  client = g_new0 (Client, 1);
+	  client->magic = CLIENT_MAGIC;
+	  client->id = NULL;
+	  client->properties = NULL;
+	  client->priority = DEFAULT_PRIORITY;
+	  client->handle = command_handle_new ((gpointer) client);
+	  client->warning = FALSE;
+	  client->get_prop_replies = NULL;
+	  client->get_prop_requests = 0;
+	  client->command_data = NULL;
+	  client->match_rule = MATCH_PROP;
+
+	  prop = (SmProp *) g_new (SmProp, 1);
+	  prop->name = g_strdup (SmRestartCommand);
+	  prop->type = g_strdup (SmLISTofARRAY8);
+	  prop->num_vals = argc;
+	  prop->vals = (SmPropValue *) g_new (SmPropValue, argc);
+
+	  for (i = 0; i < argc; i++)
+	    {
+	      prop->vals[i].length = argv[i] ? strlen (argv[i]) : 0;
+	      prop->vals[i].value = argv[i] ? g_strdup (argv[i]) : NULL;
+	    }
+
+	  APPEND (client->properties, prop);
+
+	  if (g_hash_table_lookup_extended (clients, name, (gpointer *) &hash_key, (gpointer *) &hash_client))
+	    {
+	      g_hash_table_remove (clients, name);
+	      g_free (hash_key);
+	      free_client (hash_client);
+	    }
+
+	  /* check the X-GNOME-Autostart-enabled field */
+	  if (gnome_desktop_item_attr_exists (ditem, "X-GNOME-Autostart-enabled"))
+	    {
+	      if (gnome_desktop_item_get_boolean (ditem, "X-GNOME-Autostart-enabled"))
+	        g_hash_table_insert (clients, g_strdup (name), client);
+	      else
+		free_client (client);
+	    }
+	  else
+	    g_hash_table_insert (clients, g_strdup (name), client);
+
+	  g_strfreev (argv);
+	  gnome_desktop_item_unref (ditem);
+	}
+
+      g_free (desktop_file);
+    }
+
+  g_dir_close (dir);
+}
+
+static gboolean
+hash_table_remove_cb (gpointer key, gpointer value, gpointer user_data)
+{
+  GSList **list = (GSList **) user_data;
+
+  *list = g_slist_prepend (*list, value);
+  g_free (key);
+
+  return TRUE;
+}
+
+/* read the list of system-wide autostart apps */
+static GSList *
+read_autostart_dirs (void)
+{
+  const char * const * system_config_dirs;
+  char *path;
+  gint i;
+  GHashTable *clients;
+  GSList *list = NULL;
+
+  clients = g_hash_table_new (g_str_hash, g_str_equal);
+
+  /* read directories */
+  system_config_dirs = g_get_system_config_dirs ();
+  for (i = 0; system_config_dirs[i] != NULL; i++)
+    {
+      path = g_build_filename (system_config_dirs[i], "autostart", NULL);
+      read_desktop_entries_in_dir (clients, path);
+      g_free (path);
+    }
+
+  path = g_build_filename (g_get_user_config_dir (), "autostart", NULL);
+  read_desktop_entries_in_dir (clients, path);
+  g_free (path);
+
+  /* convert the hash table into a GSList */
+  g_hash_table_foreach_remove (clients, (GHRFunc) hash_table_remove_cb, &list);
+  g_hash_table_destroy (clients);
+
+  return list;
+}
 
 /* frees lock on session at shutdown */
 void 
@@ -411,6 +623,64 @@ get_current_session(void)
   } 
 }
 
+static gboolean
+compare_restart_command (GSList *props1, GSList *props2)
+{
+  SmProp *p1, *p2;
+  GSList *node1, *node2;
+
+  for (node1 = props1; node1 != NULL; node1 = node1->next)
+    {
+      p1 = node1->data;
+      if (!strcmp (p1->name, SmRestartCommand))
+        {
+	  for (node2 = props2; node2 != NULL; node2 = node2->next)
+	    {
+	      p2 = node2->data;
+	      if (!strcmp (p2->name, SmRestartCommand))
+                {
+                  if ((p1->vals[0].value == NULL) || 
+                      (p2->vals[0].value == NULL))
+                    return FALSE;
+                  return strcmp (p1->vals[0].value, p2->vals[0].value) == 0;
+                }
+	    }
+	  break;
+        }
+    }
+
+  return FALSE;
+}
+
+static GSList *
+remove_duplicated_clients (GSList *clients)
+{
+  GSList *node;
+
+  for (node = clients; node != NULL; node = node->next)
+    {
+      GSList *sl;
+      Client *client = node->data;
+
+      for (sl = node->next; sl != NULL; sl = sl->next)
+        {
+	  Client *tmp_client = sl->data;
+	  GSList *props, *tmp_props;
+
+	  props = client->properties;
+	  tmp_props = tmp_client->properties;
+
+	  if (compare_restart_command (props, tmp_props))
+	    {
+	      REMOVE (clients, tmp_client);
+	      free_client (tmp_client);
+	      sl = node;
+	    }
+	}
+    }
+
+  return clients;
+}
 
 /* Load a session from the config file by name. */
 Session*
@@ -444,8 +714,65 @@ read_session (const char *name)
    * created from the session-properties capplet
    */
   list = read_clients (MANUAL_CONFIG_PREFIX, name, MATCH_PROP);
+  if (list) {
+    char *path;
+
+    /* migrate to newly desktop-based autostart system */
+    while (list) {
+      SmProp *prop;
+      Client *client = (Client *) list->data;
+    
+      prop = find_property_by_name (client, SmRestartCommand);
+      if (prop) {
+        GnomeDesktopItem *ditem;
+	int i;
+	char *basename, *orig_filename, *filename;
+	GString *args = g_string_new ((const char *) prop->vals[0].value);
+
+	for (i = 1; i < prop->num_vals; i++) {
+	  args = g_string_append_c (args, ' ');
+	  args = g_string_append (args, (const char *) prop->vals[i].value);
+	}
+
+	ditem = gnome_desktop_item_new ();
+	gnome_desktop_item_set_string (ditem, GNOME_DESKTOP_ITEM_EXEC, args->str);
+
+	basename = g_path_get_basename (prop->vals[0].value);
+	orig_filename = g_strdup_printf ("%s/autostart/%s.desktop", g_get_user_config_dir (), basename);
+	filename = g_strdup (orig_filename);
+	i = 2;
+	while (g_file_exists (filename)) {
+	  char *tmp = g_strdup_printf ("%s-%d", orig_filename, i);
+
+	  g_free (filename);
+	  filename = tmp;
+	  i++;
+	}
+	
+	if (!gnome_desktop_item_save (ditem, filename, TRUE, NULL))
+  	  g_warning ("Could not save %s file\n", filename);
+
+	gnome_desktop_item_unref (ditem);
+	g_string_free (args, TRUE);
+	g_free (orig_filename);
+	g_free (filename);
+      }
+
+      REMOVE (list, client);
+      free_client (client);
+    }
+
+    path = g_build_filename (g_get_home_dir (), ".gnome2/session-manual", NULL);
+    g_remove ((const char *) path);
+    g_free (path);
+  }
+
+  /* Get applications to be autostarted */
+  list = read_autostart_dirs ();
   if (list)
-	  session->client_list = g_slist_concat (session->client_list, list);
+    session->client_list = g_slist_concat (session->client_list, list);
+
+  session->client_list = remove_duplicated_clients (session->client_list);
 
   return session;
 }
